use crate::parser::template;
use crate::parser::Body;
use crate::parser::DustTag;
use crate::parser::Filter;
use crate::parser::PartialNameElement;
use crate::parser::Path;
use crate::parser::Special;
use crate::parser::Template;
use crate::parser::TemplateElement;
use crate::renderer::breadcrumb_tree::BreadcrumbTreeElement;
use crate::renderer::context_element::ContextElement;
use crate::renderer::context_element::IntoContextElement;
use crate::renderer::errors::CompileError;
use crate::renderer::errors::RenderError;
use crate::renderer::errors::WalkError;
use crate::renderer::inline_partial_tree::extract_inline_partials;
use crate::renderer::inline_partial_tree::InlinePartialTreeElement;
use crate::renderer::parameters_context::ParametersContext;
use crate::renderer::tree_walking::walk_path;
use std::borrow::Borrow;
use std::collections::HashMap;
use std::rc::Rc;

#[derive(Clone, Debug)]
pub struct DustRenderer<'a> {
    templates: HashMap<String, &'a Template<'a>>,
}

pub fn compile_template<'a>(source: &'a str) -> Result<Template<'a>, CompileError> {
    let (_remaining, parsed_template) = template(source).map_err(|err| CompileError {
        message: "Failed to compile template".to_owned(),
    })?;
    Ok(parsed_template)
}

impl<'a> DustRenderer<'a> {
    pub fn new() -> DustRenderer<'a> {
        DustRenderer {
            templates: HashMap::new(),
        }
    }

    pub fn load_source(&mut self, template: &'a Template, name: String) {
        self.templates.insert(name, template);
    }

    pub fn render<C>(&'a self, name: &str, context: Option<&C>) -> Result<String, RenderError>
    where
        C: IntoContextElement,
    {
        let breadcrumbs = context
            .map(|ctx| vec![BreadcrumbTreeElement::from_borrowed(ctx)])
            .unwrap_or(Vec::new());
        self.render_template(name, breadcrumbs.as_ref(), None)
    }

    pub fn render_template(
        &'a self,
        name: &str,
        breadcrumbs: &'a Vec<BreadcrumbTreeElement<'a>>,
        blocks: Option<&'a InlinePartialTreeElement<'a>>,
    ) -> Result<String, RenderError> {
        let main_template = match self.templates.get(name) {
            Some(tmpl) => tmpl,
            None => {
                return Err(RenderError::TemplateNotFound(name.to_owned()));
            }
        };
        let extracted_inline_partials = extract_inline_partials(main_template);
        let new_blocks = InlinePartialTreeElement::new(blocks, extracted_inline_partials);
        let new_block_context = BlockContext {
            breadcrumbs: breadcrumbs,
            blocks: &new_blocks,
        };
        self.render_body(&main_template.contents, breadcrumbs, &new_block_context)
    }

    fn render_maybe_body(
        &'a self,
        body: &'a Option<Body>,
        breadcrumbs: &'a Vec<BreadcrumbTreeElement<'a>>,
        blocks: &'a BlockContext<'a>,
    ) -> Result<String, RenderError> {
        match body {
            None => Ok("".to_owned()),
            Some(body) => Ok(self.render_body(body, breadcrumbs, blocks)?),
        }
    }

    fn render_body(
        &'a self,
        body: &'a Body,
        breadcrumbs: &'a Vec<BreadcrumbTreeElement<'a>>,
        blocks: &'a BlockContext<'a>,
    ) -> Result<String, RenderError> {
        let mut output = String::new();
        for elem in &body.elements {
            match elem {
                TemplateElement::TEIgnoredWhitespace(_) => {}
                TemplateElement::TESpan(span) => output.push_str(span.contents),
                TemplateElement::TETag(dt) => {
                    output.push_str(&self.render_tag(dt, breadcrumbs, blocks)?);
                }
            }
        }
        Ok(output)
    }

    /// For rendering a dynamic partial's name or an rvalue template
    pub fn render_partial_name(
        &'a self,
        body: &'a Vec<PartialNameElement>,
        breadcrumbs: &'a Vec<BreadcrumbTreeElement<'a>>,
    ) -> Result<String, RenderError> {
        let converted_to_template_elements: Vec<TemplateElement<'a>> =
            body.into_iter().map(|e| e.into()).collect();
        // Simple templates like partial names and reference rvalues
        // cannot contain blocks or inline partials, so we use a blank
        // BlockContext.
        let empty_block_context = BlockContext {
            breadcrumbs: &Vec::new(),
            blocks: &InlinePartialTreeElement::new(None, HashMap::new()),
        };
        self.render_body(
            &Body {
                elements: converted_to_template_elements,
            },
            breadcrumbs,
            &empty_block_context,
        )
    }

    fn render_tag(
        &'a self,
        tag: &'a DustTag,
        breadcrumbs: &'a Vec<BreadcrumbTreeElement<'a>>,
        blocks: &'a BlockContext<'a>,
    ) -> Result<String, RenderError> {
        match tag {
            DustTag::DTComment(_comment) => (),
            DustTag::DTSpecial(special) => {
                return Ok(match special {
                    Special::Space => " ",
                    Special::NewLine => "\n",
                    Special::CarriageReturn => "\r",
                    Special::LeftCurlyBrace => "{",
                    Special::RightCurlyBrace => "}",
                }
                .to_owned())
            }
            DustTag::DTLiteralStringBlock(literal) => return Ok((*literal).to_owned()),
            DustTag::DTReference(reference) => {
                let val = walk_path(breadcrumbs, &reference.path.keys)
                    .map(|ice| ice.into_context_element(self, breadcrumbs));
                match val {
                    Err(WalkError::CantWalk) | Ok(None) => return Ok("".to_owned()),
                    Ok(Some(final_val)) => {
                        return if final_val.get_context_element_reference().is_truthy() {
                            final_val
                                .get_context_element_reference()
                                .render(&Self::preprocess_filters(&reference.filters))
                        } else {
                            Ok("".to_owned())
                        };
                    }
                }
            }
            DustTag::DTSection(container) => {
                let injected_context = ParametersContext::new(self, breadcrumbs, &container.params);
                let val = walk_path(breadcrumbs, &container.path.keys)
                    .map(|ice| ice.into_context_element(self, breadcrumbs));
                match val {
                    Err(WalkError::CantWalk) => {
                        // TODO
                        let new_breadcrumbs = self.new_breadcrumbs_section(
                            breadcrumbs,
                            None,
                            Some(&injected_context),
                            &container.explicit_context,
                            None,
                        );
                    }
                    Ok(final_val) => {
                        // TODO
                    }
                }
            }
            _ => panic!("Unsupported tag"),
        }
        Ok("".to_owned())
    }

    fn new_breadcrumbs_section<'b>(
        &'b self,
        breadcrumbs: &'b Vec<BreadcrumbTreeElement<'b>>,
        index_context: Option<&'b dyn IntoContextElement>,
        injected_context: Option<&'b dyn IntoContextElement>,
        explicit_context: &Option<Path<'b>>,
        new_context_element: Option<&'b dyn ContextElement>,
    ) -> Option<Vec<BreadcrumbTreeElement<'b>>> {
        // If none of the additional contexts are present, return None
        // to signal that the original breadcrumbs should be used
        // rather than incurring a copy here.
        match (
            index_context,
            injected_context,
            explicit_context,
            new_context_element,
        ) {
            (None, None, None, None) => return None,
            _ => (),
        }

        // If there is an explicit context, then drop all the current
        // context
        let mut new_stack = match explicit_context {
            Some(_) => Vec::with_capacity(4),
            None => breadcrumbs.clone(),
        };

        explicit_context.as_ref().map(|path| {
            walk_path(breadcrumbs, &path.keys)
                .map(|ice| ice.into_context_element(self, breadcrumbs))
                .ok()
                .flatten()
                .map(|val| {
                    if val.get_context_element_reference().is_truthy() {
                        new_stack.push(std::convert::From::from(val))
                    }
                });
        });
        injected_context.map(|ctx| new_stack.push(BreadcrumbTreeElement::from_borrowed(ctx)));
        new_context_element.map(|ctx| {
            new_stack.push(BreadcrumbTreeElement::from_borrowed(
                ctx.from_context_element(),
            ))
        });
        index_context.map(|ctx| new_stack.push(BreadcrumbTreeElement::from_borrowed(ctx)));

        Some(new_stack)
    }

    fn new_breadcrumbs_partial<'b>(
        &'b self,
        maybe_breadcrumbs: &'b Vec<BreadcrumbTreeElement<'b>>,
        explicit_context_maybe_breadcrumbs: &'a Vec<BreadcrumbTreeElement<'a>>,
        injected_context: Option<&'b dyn IntoContextElement>,
        explicit_context: &Option<Path<'b>>,
    ) {
        /*
        // If none of the additional contexts are present, return None
        // to signal that the original breadcrumbs should be used
        // rather than incurring a copy here.
        match (injected_context, explicit_context) {
            (None, None) => return None,
            _ => (),
        };

        // If there is an explicit context, then drop all the current
        // context
        let mut parent = match explicit_context {
            Some(_) => None,
            None => maybe_breadcrumbs,
        };
        let mut new_nodes: Vec<BreadcrumbTreeElement> = Vec::new();

        injected_context.map(|ctx| {
            // Special case: when there is no explicit context, the
            // injected context gets inserted 1 spot behind the
            // current context. Otherwise, the injected context gets
            // added after the current context but before the explicit
            // context.
            match explicit_context {
                None => {
                    let (new_parent, passed_nodes) =
                        Self::split_tree_at_predicate(parent, |b| b.get_ice().is_pseudo_element());
                    parent = new_parent;
                    new_nodes.extend(passed_nodes.iter().map(|b| b.get_element().clone()));
                }
                _ => new_nodes.push(BreadcrumbTreeElement::from_borrowed(ctx)),
            }
        });
        explicit_context.as_ref().map(|path| {
            // TODO: should resolving the value here use
            // explicit_context_maybe_breadcrumbs or
            // maybe_breadcrumbs?
            walk_path(explicit_context_maybe_breadcrumbs, &path.keys)
                .map(|ice| ice.into_context_element(self, maybe_breadcrumbs))
                .ok()
                .flatten()
                .map(|val| {
                    if val.get_context_element_reference().is_truthy() {
                        new_nodes.push(std::convert::From::from(val));
                    }
                });
        });

        Some((parent, new_nodes))*/
        todo!()
    }

    fn preprocess_filters(filters: &Vec<Filter>) -> Vec<Filter> {
        let mut final_filters: Vec<Filter> = filters
            .into_iter()
            .filter(|f| f != &&Filter::DisableHtmlEncode)
            .map(|f| f.clone())
            .collect();

        // If the user has not specified any escaping filter (|s or
        // |h), automatically add an html escape filter
        if !filters.iter().any(|f| f == &Filter::DisableHtmlEncode) {
            final_filters.push(Filter::HtmlEncode);
        }
        final_filters
    }
}

struct BlockContext<'a> {
    /// The breadcrumbs at the time of entering the current partial
    breadcrumbs: &'a Vec<BreadcrumbTreeElement<'a>>,
    blocks: &'a InlinePartialTreeElement<'a>,
}
